#!/usr/bin/env python3
"""
Network Automation Tool
Parameters:
    - IP: Target device IP
    - port: SSH port
Additional parameters in comments:
    - username: SSH username
    - password: SSH password
    - commands: List of commands to execute
"""
import paramiko
import netmiko
import argparse
import logging
from datetime import datetime
import json
from pathlib import Path
import threading
import queue
import time
import yaml
import jinja2
import re

class NetworkAutomation:
    def __init__(self, config_file=None):
        self.config = self.load_config(config_file) if config_file else {}
        self.results_queue = queue.Queue()
        self.setup_logging()
        
    def setup_logging(self):
        """Set up logging configuration"""
        log_dir = Path("automation_logs")
        log_dir.mkdir(exist_ok=True)
        
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s',
            handlers=[
                logging.FileHandler(log_dir / "automation.log"),
                logging.StreamHandler()
            ]
        )
        
    def load_config(self, config_file):
        """Load configuration from YAML file"""
        with open(config_file, 'r') as f:
            return yaml.safe_load(f)
            
    def render_template(self, template_file, variables):
        """Render Jinja2 template"""
        template_dir = Path("templates")
        env = jinja2.Environment(loader=jinja2.FileSystemLoader(str(template_dir)))
        template = env.get_template(template_file)
        return template.render(variables)
        
    def connect_device(self, device):
        """Connect to network device"""
        try:
            # Determine device type for netmiko
            device_type = device.get('device_type', 'cisco_ios')
            
            connection = netmiko.ConnectHandler(
                device_type=device_type,
                ip=device['ip'],
                username=device['username'],
                password=device['password'],
                port=device.get('port', 22),
                secret=device.get('enable_password'),
                verbose=False
            )
            
            if device.get('enable_password'):
                connection.enable()
                
            return connection
            
        except Exception as e:
            logging.error(f"Error connecting to {device['ip']}: {str(e)}")
            return None
            
    def execute_commands(self, connection, commands, device_ip):
        """Execute commands on device"""
        results = []
        
        try:
            for command in commands:
                start_time = time.time()
                output = connection.send_command(command)
                duration = time.time() - start_time
                
                results.append({
                    'command': command,
                    'output': output,
                    'duration': duration,
                    'timestamp': datetime.now().isoformat()
                })
                
        except Exception as e:
            logging.error(f"Error executing commands on {device_ip}: {str(e)}")
            
        return results
        
    def configure_device(self, connection, config_commands, device_ip):
        """Configure device with commands"""
        try:
            output = connection.send_config_set(config_commands)
            connection.save_config()
            return output
            
        except Exception as e:
            logging.error(f"Error configuring {device_ip}: {str(e)}")
            return None
            
    def backup_config(self, connection, device_ip):
        """Backup device configuration"""
        try:
            config = connection.send_command("show running-config")
            
            # Save config to file
            backup_dir = Path("config_backups")
            backup_dir.mkdir(exist_ok=True)
            
            filename = backup_dir / f"{device_ip}_{datetime.now().strftime('%Y%m%d_%H%M%S')}.cfg"
            with open(filename, 'w') as f:
                f.write(config)
                
            return str(filename)
            
        except Exception as e:
            logging.error(f"Error backing up config for {device_ip}: {str(e)}")
            return None
            
    def verify_config(self, connection, verification_commands):
        """Verify configuration changes"""
        results = []
        
        try:
            for command in verification_commands:
                output = connection.send_command(command)
                results.append({
                    'command': command,
                    'output': output,
                    'timestamp': datetime.now().isoformat()
                })
                
        except Exception as e:
            logging.error(f"Error verifying configuration: {str(e)}")
            
        return results
        
    def parse_output(self, output, template):
        """Parse command output using TextFSM template"""
        try:
            template_dir = Path("templates")
            template_file = template_dir / template
            
            with open(template_file, 'r') as f:
                fsm = textfsm.TextFSM(f)
                result = fsm.ParseText(output)
                
            return [dict(zip(fsm.header, row)) for row in result]
            
        except Exception as e:
            logging.error(f"Error parsing output: {str(e)}")
            return None
            
    def run_automation(self, devices=None, commands=None, config_commands=None,
                    verification_commands=None):
        """Run automation tasks on devices"""
        if devices is None:
            devices = self.config.get('devices', [])
            
        if commands is None:
            commands = self.config.get('commands', [])
            
        results = []
        
        for device in devices:
            device_result = {
                'device': device['ip'],
                'timestamp': datetime.now().isoformat(),
                'status': 'unknown'
            }
            
            # Connect to device
            connection = self.connect_device(device)
            if not connection:
                device_result['status'] = 'connection_failed'
                results.append(device_result)
                continue
                
            try:
                # Backup existing config
                backup_file = self.backup_config(connection, device['ip'])
                device_result['backup_file'] = backup_file
                
                # Execute show commands
                if commands:
                    device_result['command_output'] = self.execute_commands(
                        connection, commands, device['ip']
                    )
                    
                # Apply configuration changes
                if config_commands:
                    device_result['config_output'] = self.configure_device(
                        connection, config_commands, device['ip']
                    )
                    
                # Verify changes
                if verification_commands:
                    device_result['verification_output'] = self.verify_config(
                        connection, verification_commands
                    )
                    
                device_result['status'] = 'success'
                
            except Exception as e:
                device_result['status'] = 'error'
                device_result['error'] = str(e)
                logging.error(f"Error on device {device['ip']}: {str(e)}")
                
            finally:
                connection.disconnect()
                results.append(device_result)
                
        # Save results
        self.save_results(results)
        
    def save_results(self, results):
        """Save automation results"""
        output_dir = Path("automation_results")
        output_dir.mkdir(exist_ok=True)
        
        filename = output_dir / f"results_{datetime.now().strftime('%Y%m%d_%H%M%S')}.json"
        with open(filename, 'w') as f:
            json.dump(results, f, indent=2)
            
        logging.info(f"Results saved to {filename}")
        self.print_summary(results)
        
    def print_summary(self, results):
        """Print automation summary"""
        print("\nAutomation Summary")
        print("=" * 50)
        print(f"Total devices: {len(results)}")
        print(f"Successful: {len([r for r in results if r['status'] == 'success'])}")
        print(f"Failed: {len([r for r in results if r['status'] != 'success'])}")
        print("\nDevice Details:")
        
        for result in results:
            print(f"\nDevice: {result['device']}")
            print(f"Status: {result['status']}")
            if result.get('backup_file'):
                print(f"Backup saved: {result['backup_file']}")
            if result.get('error'):
                print(f"Error: {result['error']}")

def create_template_structure():
    """Create template directory structure"""
    template_dir = Path("templates")
    template_dir.mkdir(exist_ok=True)
    
    # Create example templates
    templates = {
        'cisco_interface.j2': """interface {{ interface }}
 description {{ description }}
 ip address {{ ip_address }} {{ subnet_mask }}
 {% if shutdown %}shutdown{% else %}no shutdown{% endif %}""",
        
        'cisco_vlan.j2': """vlan {{ vlan_id }}
 name {{ vlan_name }}""",
        
        'cisco_acl.j2': """ip access-list extended {{ acl_name }}
{% for rule in rules %}
 {{ rule.action }} {{ rule.protocol }} {{ rule.source }} {{ rule.destination }}
{% endfor %}"""
    }
    
    for name, content in templates.items():
        template_file = template_dir / name
        if not template_file.exists():
            with open(template_file, 'w') as f:
                f.write(content)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Network Automation Tool")
    parser.add_argument("--config", help="YAML configuration file")
    parser.add_argument("--device", help="Single device IP address")
    parser.add_argument("--commands", help="Comma-separated list of commands")
    parser.add_argument("--template", help="Use template for configuration")
    args = parser.parse_args()
    
    # Create template structure if it doesn't exist
    create_template_structure()
    
    automation = NetworkAutomation(args.config)
    
    if args.device and args.commands:
        # Single device mode
        devices = [{
            'ip': args.device,
            'username': input("Username: "),
            'password': getpass.getpass("Password: "),
            'device_type': 'cisco_ios'  # Default to Cisco IOS
        }]
        commands = args.commands.split(',')
        automation.run_automation(devices=devices, commands=commands)
    elif args.config:
        # Configuration file mode
        automation.run_automation()
    else:
        parser.error("Either --config or both --device and --commands must be specified")
